unit Main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls, Menus, Grids, ESRI;

var
  BG_COLOR : dword = clSkyBlue;

type
  TFormSHPViewer = class(TForm)
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Open1: TMenuItem;
    Close1: TMenuItem;
    N1: TMenuItem;
    Exit1: TMenuItem;
    GroupBox1: TGroupBox;
    Splitter1: TSplitter;
    GroupBox2: TGroupBox;
    grd: TDrawGrid;
    img: TPaintBox;
    OpenDlg: TOpenDialog;
    procedure FormActivate(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure Close1Click(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
    procedure Open1Click(Sender: TObject);
    procedure grdDrawCell(Sender: TObject; ACol, ARow: Integer; Rect: TRect; State: TGridDrawState);
    procedure imgPaint(Sender: TObject);
    procedure SetCenter;
    function RectInView(r1,r2 : TRect) : boolean;
    function PointInView(Pnt : TPoint; r2 : TRect) : boolean;
    procedure Splitter1Moved(Sender: TObject);
    procedure FormMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure imgMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure imgMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure imgMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure grdMouseWheelDown(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
    procedure grdMouseWheelUp(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
  private
    SHP      : TESRIDataSet;
    Zoom     : double;
    Center   : TPoint;
    Current  : TPoint;
    Resizing : boolean;
    Ratio    : double;
    CBmp     : TBitmap;
    Moved    : boolean;
    function OpenShapes(FileName : string) : boolean;
    procedure SetupGrid;
    procedure WMStartResize(var Msg: TMessage); message WM_ENTERSIZEMOVE;
    procedure WMEndResize(var Msg: TMessage); message WM_EXITSIZEMOVE;
    function GetBoundRect : TRect;
    function CorrectPoint(SrcPnt : TPoint; BoundRect : TRect) : TPoint;
    function PntToLL(SrcPnt : TPoint; BoundRect : TRect) : TPoint; overload;
    function PntToLL(SrcX,SrcY : integer; BoundRect : TRect) : TPoint; overload;
    procedure RecreateCBmp;
  end;

var
  FormSHPViewer: TFormSHPViewer;

implementation

{$R *.dfm}

procedure TFormSHPViewer.Close1Click(Sender: TObject);
begin
  if assigned(SHP) then FreeAndNil(SHP);
  grd.ColCount := 0;
  grd.RowCount := 1;
  grd.FixedRows := 0;
end;

procedure TFormSHPViewer.Exit1Click(Sender: TObject);
begin
  Close;
end;

procedure TFormSHPViewer.FormActivate(Sender: TObject);
begin
  if Paramcount > 0 then OpenShapes(Paramstr(1));
  if not assigned(CBmp) then
  begin
    CBmp := TBitmap.Create;
    CBmp.Width := img.ClientWidth;
    CBmp.Height := img.ClientHeight;
    CBmp.PixelFormat := pf32bit;
  end;
end;

procedure TFormSHPViewer.FormCreate(Sender: TObject);
begin
  SHP := nil;
  Resizing := false;
  CBmp := nil;
end;

procedure TFormSHPViewer.FormMouseWheel(Sender: TObject; Shift: TShiftState;
  WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
var
  pnt : TPoint;
  ll  : TPoint;
  kk  : double;
  rr  : TRect;
begin
  if not assigned(SHP) or ((Zoom < 0.000002) and (WheelDelta < 0)) or ((Zoom > 0.1) and (WheelDelta > 0)) then exit;
  GetCursorPos(pnt);
  pnt := img.ScreenToClient(pnt);
  if WheelDelta > 0 then kk := 0.7 else kk := 1.43;
  ll := PntToLL(pnt,GetBoundRect);
  Center.X := trunc(ll.X + (Center.X - ll.X) * kk);
  Center.Y := trunc(ll.Y - (ll.Y - Center.Y) * kk);
  Zoom := Zoom / kk;
  img.Invalidate;
  Handled := true;
end;

function TFormSHPViewer.PntToLL(SrcPnt : TPoint; BoundRect : TRect) : TPoint;
begin
  result.X := trunc(BoundRect.Left + (BoundRect.Right - BoundRect.Left) * (SrcPnt.X / img.ClientWidth));
  result.Y := trunc(BoundRect.Bottom + (BoundRect.Top - BoundRect.Bottom) * (1 - (SrcPnt.Y / img.ClientHeight)));
end;

function TFormSHPViewer.PntToLL(SrcX,SrcY : integer; BoundRect : TRect) : TPoint;
begin
  result.X := trunc(BoundRect.Left + (BoundRect.Right - BoundRect.Left) * (SrcX / img.ClientWidth));
  result.Y := trunc(BoundRect.Bottom + (BoundRect.Top - BoundRect.Bottom) * (1 - (SrcY / img.ClientHeight)));
end;

function TFormSHPViewer.GetBoundRect : TRect;
begin
  result.Left := trunc(Center.X - (img.ClientWidth div 2) / Zoom);
  result.Right := trunc(Center.X + (img.ClientWidth div 2) / Zoom);
  result.Top := trunc(Center.Y + (img.ClientHeight div 2) / Zoom);
  result.Bottom := trunc(Center.Y - (img.ClientHeight div 2) / Zoom);
end;

procedure TFormSHPViewer.WMStartResize(var Msg: TMessage);
begin
  Ratio := GroupBox1.Height / ClientHeight;
  Resizing := true;
end;

procedure TFormSHPViewer.RecreateCBmp;
begin
  if assigned(CBmp) then
  begin
    CBmp.Free;
    CBmp := TBitmap.Create;
    CBmp.Width := img.ClientWidth;
    CBmp.Height := img.ClientHeight;
    CBmp.PixelFormat := pf32bit;
  end;
end;

procedure TFormSHPViewer.WMEndResize(var Msg: TMessage);
begin
  RecreateCBmp;
  Resizing := false;
  if not Visible then exit;
  if assigned(SHP) then
  begin
    if SHP.DBF then
      GroupBox1.Height := trunc(ClientHeight * Ratio)
    else
      GroupBox1.Height := ClientHeight;
    SetCenter;
  end;
  img.Invalidate;
end;

procedure TFormSHPViewer.grdDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
var
  sVal : string;
begin
  if not assigned(SHP) then exit;
  sVal := '';
  if ARow = 0 then
    sVal := SHP.Attributes[ACol].Name
  else
    sVal := SHP.Attributes[ACol].Values[ARow-1].SValue;
  grd.Canvas.Lock;
  try
    grd.Canvas.Font.Charset := SHP.CodePage;
    grd.Canvas.Font.Size := 10;
    grd.Canvas.TextRect(Rect,Rect.Left+1,Rect.Top+1,sVal);
  finally
    grd.Canvas.Unlock;
  end;
end;

procedure TFormSHPViewer.grdMouseWheelDown(Sender: TObject; Shift: TShiftState;
  MousePos: TPoint; var Handled: Boolean);
var
  pnt : TPoint;
begin
  GetCursorPos(pnt);
  pnt := grd.ScreenToClient(pnt);
  if (pnt.X < 0) or (pnt.X > ClientWidth) or (pnt.Y < 0) or (pnt.Y > ClientHeight) then
    FormSHPViewer.OnMouseWheel(Sender,Shift,-1,MousePos,Handled);
end;

procedure TFormSHPViewer.grdMouseWheelUp(Sender: TObject; Shift: TShiftState;
  MousePos: TPoint; var Handled: Boolean);
var
  pnt : TPoint;
begin
  GetCursorPos(pnt);
  pnt := grd.ScreenToClient(pnt);
  if (pnt.X < 0) or (pnt.X > ClientWidth) or (pnt.Y < 0) or (pnt.Y > ClientHeight) then
    FormSHPViewer.OnMouseWheel(Sender,Shift,1,MousePos,Handled);
end;

procedure TFormSHPViewer.Open1Click(Sender: TObject);
begin
  if OpenDlg.Execute then OpenShapes(OpenDlg.FileName);
end;

function TFormSHPViewer.OpenShapes(FileName : string) : boolean;
begin
  result := false;
  if not FileExists(FileName) then
  begin
    ShowMessage('File "' + FileName + '" does not exist');
    exit;
  end;
  if assigned(SHP) then SHP.Free;
  SHP := TESRIDataSet.Create(FileName);
  if SHP.DBF then
  begin
    GroupBox2.Visible := true;
    GroupBox1.Height := trunc(ClientHeight * 0.7);
    SetupGrid;
  end
  else
  begin
    GroupBox2.Visible := false;
    GroupBox1.Height := ClientHeight;
  end;
  RecreateCBmp;
  if assigned(SHP.Shapes) and (SHP.Shapes.Count > 0) then
    Caption := 'SHPViewer - [' + FileName + '] [Shape File Type: Polygons]'
  else
  if assigned(SHP.Lines) and (SHP.Lines.Count > 0) then
    Caption := 'SHPViewer - [' + FileName + '] [Shape File Type: Lines]'
  else
  if assigned(SHP.Points) and (SHP.Points.Count > 0) then
    Caption := 'SHPViewer - [' + FileName + '] [Shape File Type: Points]'
  else
    Caption := 'SHPViewer - [' + FileName + '] [Shape File Type: Unknown]';
  SetCenter;
  grd.Invalidate;
  img.Invalidate;
  result := true;
end;

procedure TFormSHPViewer.SetupGrid;
begin
  if not assigned(SHP) then exit;
  grd.ColCount := SHP.Attributes.Count;
  if assigned(SHP.Shapes) and (SHP.Shapes.Count > 0) then
    grd.RowCount := SHP.Shapes.Count
  else
  if assigned(SHP.Lines) and (SHP.Lines.Count > 0) then
    grd.RowCount := SHP.Lines.Count
  else
  if assigned(SHP.Points) and (SHP.Points.Count > 0) then
    grd.RowCount := SHP.Points.Count
  else
    grd.RowCount := 0;
  if grd.RowCount > 1 then grd.FixedRows := 1;
end;

procedure TFormSHPViewer.Splitter1Moved(Sender: TObject);
begin
  RecreateCBmp;
  SetCenter;
  img.Invalidate;
end;

procedure TFormSHPViewer.SetCenter;
var
  ii : integer;
  rr : TRect;
  zx,zy : double;
begin
  if not assigned(SHP) then exit;
  if assigned(SHP.Shapes) and (SHP.Shapes.Count > 0) then
  begin
    rr := SHP.Shapes[0].Rect;
    for ii:=1 to SHP.Shapes.Count-1 do
    begin
      if rr.Left > SHP.Shapes[ii].Rect.Left then rr.Left := SHP.Shapes[ii].Rect.Left;
      if rr.Right < SHP.Shapes[ii].Rect.Right then rr.Right := SHP.Shapes[ii].Rect.Right;
      if rr.Top < SHP.Shapes[ii].Rect.Top then rr.Top := SHP.Shapes[ii].Rect.Top;
      if rr.Bottom > SHP.Shapes[ii].Rect.Bottom then rr.Bottom := SHP.Shapes[ii].Rect.Bottom;
    end;
  end
  else
  if assigned(SHP.Lines) and (SHP.Lines.Count > 0) then
  begin
    rr := SHP.Lines[0].Rect;
    for ii:=1 to SHP.Lines.Count-1 do
    begin
      if rr.Left > SHP.Lines[ii].Rect.Left then rr.Left := SHP.Lines[ii].Rect.Left;
      if rr.Right < SHP.Lines[ii].Rect.Right then rr.Right := SHP.Lines[ii].Rect.Right;
      if rr.Top < SHP.Lines[ii].Rect.Top then rr.Top := SHP.Lines[ii].Rect.Top;
      if rr.Bottom > SHP.Lines[ii].Rect.Bottom then rr.Bottom := SHP.Lines[ii].Rect.Bottom;
    end;
  end
  else
  if assigned(SHP.Points) and (SHP.Points.Count > 0) then
  begin
    rr.Left := SHP.Points[0].X;
    rr.Right := SHP.Points[0].X;
    rr.Top := SHP.Points[0].Y;
    rr.Bottom := SHP.Points[0].Y;
    for ii:=1 to SHP.Points.Count-1 do
    begin
      if rr.Left > SHP.Points[ii].X then rr.Left := SHP.Points[ii].X;
      if rr.Right < SHP.Points[ii].X then rr.Right := SHP.Points[ii].X;
      if rr.Top < SHP.Points[ii].Y then rr.Top := SHP.Points[ii].Y;
      if rr.Bottom > SHP.Points[ii].Y then rr.Bottom := SHP.Points[ii].Y;
    end;
  end
  else
    exit;
  Center.X := (rr.Left + rr.Right) div 2;
  Center.Y := (rr.Top + rr.Bottom) div 2;
  zx := img.ClientWidth / (rr.Right - rr.Left);
  zy := img.ClientHeight / (rr.Top - rr.Bottom);
  if zx < zy then
    Zoom := zx
  else
    Zoom := zy;
end;

function TFormSHPViewer.RectInView(r1,r2 : TRect) : boolean;
begin
  result := not ((r1.Left > r2.Right) or (r1.Right < r2.Left) or (r1.Top < r2.Bottom) or (r1.Bottom > r2.Top));
end;

function TFormSHPViewer.PointInView(Pnt : TPoint; r2 : TRect) : boolean;
begin
  result := (pnt.X >= r2.Left) and (pnt.X <= r2.Right) and (pnt.Y >= r2.Bottom) and (pnt.Y <= r2.Top);
end;

function TFormSHPViewer.CorrectPoint(SrcPnt : TPoint; BoundRect : TRect) : TPoint;
begin
  result.X := trunc((SrcPnt.X - BoundRect.Left) * Zoom);
  result.Y := img.ClientHeight - trunc((SrcPnt.Y - BoundRect.Bottom) * Zoom);
end;

procedure TFormSHPViewer.imgMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if not assigned(SHP) or (Button <> mbLeft) then exit;
  Current.X := X;
  Current.Y := Y;
  Moved := false;
end;

procedure TFormSHPViewer.imgMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  bmp : TBitmap;
begin
  if not assigned(SHP) or not assigned(CBmp) or not (ssLeft in Shift) then exit;
  Moved := true;
  bmp := TBitmap.Create;
  try
    bmp.Width := img.ClientWidth;
    bmp.Height := img.ClientHeight;
    bmp.PixelFormat := pf32bit;
    bmp.Canvas.Pen.Color := clBlack;
    bmp.Canvas.Brush.Color := BG_COLOR;
    bmp.Canvas.FillRect(img.ClientRect);
    bmp.Canvas.Draw(X-Current.X,Y-Current.Y,CBmp);
    img.Canvas.Draw(0,0,bmp);
  finally
    bmp.Free;
  end;
end;

procedure TFormSHPViewer.imgMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  pnt1,pnt2 : TPoint;
begin
  if not assigned(SHP) or not assigned(CBmp) or not Moved then exit;
  Moved := false;
  pnt1 := PntToLL(Current,GetBoundRect);
  pnt2 := PntToLL(X,Y,GetBoundRect);
  Center.X := Center.X - (pnt2.X - pnt1.X);
  Center.Y := Center.Y - (pnt2.Y - pnt1.Y);
  img.Invalidate;
end;

procedure TFormSHPViewer.imgPaint(Sender: TObject);
var
  cc : TCanvas;
  BoundRect,rr : TRect;
  iShape, iPolygon : integer;
  pnts : array of TPoint;
  pnt : TPoint;

  procedure GetPoints(SrcPoints : array of TPoint; StartPoint, PointCount : integer);
  var
    ii : integer;
    cntr : integer;
  begin
    setlength(pnts,PointCount);
    cntr := 1;
    pnts[0] := CorrectPoint(SrcPoints[StartPoint],BoundRect);
    for ii:=StartPoint+1 to (StartPoint+PointCount-1) do
    begin
      pnts[cntr] := CorrectPoint(SrcPoints[ii],BoundRect);
      if (pnts[cntr].X <> pnts[cntr-1].X) or (pnts[cntr].Y <> pnts[cntr-1].Y) then inc(cntr);
    end;
    setlength(pnts,cntr-1);
  end;

begin
  if Resizing or not assigned(CBmp) then exit;
  cc := CBmp.Canvas;
  cc.Lock;
  try
    cc.Pen.Color := clBlack;
    cc.Brush.Color := BG_COLOR;
    cc.FillRect(img.ClientRect);
    if not assigned(SHP) then exit;
    BoundRect := GetBoundRect;
    cc.Pen.Color := clBlack;
    cc.Brush.Color := clWhite;
    if assigned(SHP.Shapes) and (SHP.Shapes.Count > 0) then
    begin
      for iShape:=0 to SHP.Shapes.Count-1 do
      begin
        if RectInView(SHP.Shapes[iShape].Rect,BoundRect) then
        begin
          for iPolygon:=0 to length(SHP.Shapes[iShape].Parts)-1 do
          begin
            if iPolygon = (length(SHP.Shapes[iShape].Parts)-1) then
              GetPoints(SHP.Shapes[iShape].Points, SHP.Shapes[iShape].Parts[iPolygon],
                length(SHP.Shapes[iShape].Points) - SHP.Shapes[iShape].Parts[iPolygon] + 1)
            else
              GetPoints(SHP.Shapes[iShape].Points, SHP.Shapes[iShape].Parts[iPolygon],
                SHP.Shapes[iShape].Parts[iPolygon+1] - SHP.Shapes[iShape].Parts[iPolygon] + 1);
            cc.Polygon(pnts);
          end;
        end;
      end;
    end
    else
    if assigned(SHP.Lines) and (SHP.Lines.Count > 0) then
    begin
      for iShape:=0 to SHP.Lines.Count-1 do
      begin
        if RectInView(SHP.Lines[iShape].Rect,BoundRect) then
        begin
          for iPolygon:=0 to length(SHP.Lines[iShape].Parts)-1 do
          begin
            if iPolygon = (length(SHP.Lines[iShape].Parts)-1) then
              GetPoints(SHP.Lines[iShape].Points, SHP.Lines[iShape].Parts[iPolygon],
                length(SHP.Lines[iShape].Points) - SHP.Lines[iShape].Parts[iPolygon] + 1)
            else
              GetPoints(SHP.Lines[iShape].Points, SHP.Lines[iShape].Parts[iPolygon],
                SHP.Lines[iShape].Parts[iPolygon+1] - SHP.Lines[iShape].Parts[iPolygon] + 1);
            cc.Polyline(pnts);
          end;
        end;
      end;
    end
    else
    if assigned(SHP.Points) and (SHP.Points.Count > 0) then
    begin
      for iShape:=0 to SHP.Points.Count-1 do
      begin
        pnt.X := SHP.Points[iShape].X;
        pnt.Y := SHP.Points[iShape].Y;
        if PointInView(pnt,BoundRect) then
        begin
          pnt := CorrectPoint(pnt,BoundRect);
          rr.Left := pnt.X-2;
          rr.Right := pnt.X+2;
          rr.Top := pnt.Y+2;
          rr.Bottom := pnt.Y-2;
          cc.Ellipse(rr);
        end;
      end;
    end
    else
      exit;
  finally
    cc.Unlock;
    img.Canvas.Draw(0,0,CBmp);
    img.Canvas.Brush.Color := clBlack;
    setlength(pnts,5);
    img.Canvas.FrameRect(img.ClientRect);
  end;
end;

end.
